/*
 * Readings from encoder testing 
 * left was -618, right -580
 * left - right = -38
 * -38 * -50 / 100 = 19
 * Validating the following balance algorithm
 * const int16_t DISTANCE_DIFF_RESPONSE = -50;
 *   motorSpeed = constrain (motorSpeed, -MOTOR_SPEED_LIMIT, MOTOR_SPEED_LIMIT);
 * int16_t distanceDiff = distance_l - distance_r;
 * motors.setSpeeds(
 *    motorSpeed + distanceDiff * DISTANCE_DIFF_RESPONSE / 100,
 *    motorSpeed - distanceDiff * DISTANCE_DIFF_RESPONSE / 100);
 * #if <statement>
 * #elif
 * #else
 * #endif
 * 75.81 x 2.88 x 12 = 2620 counts / rev
 * 80 x PI = 251 mm / rev
 * 2620 / 251 = 10 cnt / mm
 * 2 cnt / ms x 0.1 mm / cnt = 0.2 mm / ms 
 * or 200 mm / s
*/
#include <Balboa32U4.h>
#include <Wire.h>
#include <LSM6.h>
#include <LIS3MDL.h>
#include "Balance.h"

LSM6 imu;
LIS3MDL mag;
Balboa32U4Motors motors;
Balboa32U4Encoders encoders;
Balboa32U4ButtonA buttonA;
Balboa32U4ButtonB buttonB;
Balboa32U4ButtonC buttonC;
QTRSensors lineSensors;

#define CAL false
#define LINE false
#define LnLED false

const double rad_to_deg = 180 / PI; // 57296 
const double mag_scale = 6842;
LIS3MDL::vector<int16_t> mag_min = {-7424, -338, -3216};
LIS3MDL::vector<int16_t> mag_max = {-6566, 2290, -18};
LIS3MDL::vector<int16_t> mag_off = {-6995, 976, -1617};
 

double hdg;

void setup() {
  Wire.begin ();
  Serial1.begin (9600);

  mag.init ();
  mag.enableDefault ();

  ledYellow(false);
  ledRed(true);
  balanceSetup();
  ledRed(false);
}

void loop() {
  static bool enableDrive = false;
  static bool enableLine = false;
  
  balanceUpdate ();
  if (isBalancing ()) {
    if (enableDrive) {
      Test_Track ();
    }
    if (enableLine) {
      #if LINE
      #if LnLED
      uint8_t tmp = Read_LineSensor ();
      #else
      uint16_t tmp = Read_LineSensor ();
      #endif
      #else
      Motion_Test ();
      #endif
    }
  }
  else {
    balanceDrive (0, 0);
    if (buttonA.getSingleDebouncedPress()) {
      enableDrive = false;
      enableLine = true;
      standUp ();
    }
    else if (buttonB.getSingleDebouncedPress()) {
      enableDrive = false;
      enableLine = false;
      standUp ();
    }
    else if (buttonC.getSingleDebouncedPress()) {
      enableDrive = true;
      enableLine = false;
      standUp ();
    }
  }
  ledRed (balanceUpdateDelayed ());
  int32_t fallingAngleOffset = angleRate * ANGLE_RATE_RATIO - angle;
  if (fallingAngleOffset > 0) {
    ledYellow (true);
    ledGreen (false);
  }
  else {
    ledYellow (false);
    ledGreen (true);
  }
}

/*
* Line sensor program weighs each sensor
* for robot position on the line.
* 
* The central detectors(S2,S3&S4) are
* spaced close enough to detect the
* line(i.e 0.375"[9.5mm]). 
* 
* Tolerance of line detection works best
* with a limit value approximately 1000
* in my line detection process.
* 
*/
#if LnLED
uint8_t Read_LineSensor () {
  uint8_t lnLeds;
  
  // record number of contributing
  // sensors.
  lnLeds = 0;
  for (uint8_t i = 0; i < SensorCount; i++) {
    if (i > 0) lnLeds = lnLeds << 1;
    if (sensorValues [i] > 1000) lnLeds += 1;
  }
  #if LINE
  if (lnLeds < 16) Serial1.print ("0X0");
  else Serial1.print ("0X");
  Serial1.println (lnLeds, HEX);
  delay (1000);
  #endif
  return lnLeds;
}
#else
int16_t Read_LineSensor () {
  const uint8_t STEP = 20;
  const uint16_t SetPoint = STEP * (SensorCount - 1) / 2;
  uint8_t S;
  int16_t lnpos;
  
  // record number of contributing
  // sensors.
  S = 0;
  lnpos = 0;
  for (uint8_t i = 0; i < SensorCount; i++) {
    if (sensorValues [i] > 1000) {
      lnpos += i * STEP;
      S++;
    }
  }
  /*
  * Since the detection gap of the outside
  * (S5,S1) and inner detectors(S2,S3&S4) is
  * 0.75inch(19mm); track width based on
  * electrical tape; ~0.5inch(13mm); can
  * evade detection.
  * 
  * Balboa line sensor assembly has been
  * replaced by Pololu QTR RC sensor
  * array with six equidistance detctors.
  * 
  */ 
  if (S == 0) {
    lnpos = -STEP;
  }
  else if (S == SensorCount) {
    lnpos = STEP * SensorCount;
  }
  else {
    lnpos /= S;
  }
  lnpos -= SetPoint;
  #if LINE
  Serial1.println (lnpos);
  delay (1000);
  #endif
  return lnpos;
}
#endif

void Read_Mag () {
  float az;
  double mgX, mgY, mgZ;

  mag.read ();

  #if CAL
  static bool initial;
  if (initial == false) {
    mag_min = {32767, 32767, 32767};
    mag_max = {-32768, -32768, -32768};
    mag_off = {0, 0, 0};
    initial = true;
  }
  else {
    mag_min.x = min (mag_min.x, mag.m.x);
    mag_min.y = min (mag_min.y, mag.m.y);
    mag_min.z = min (mag_min.z, mag.m.z);

    mag_max.x = max (mag_max.x, mag.m.x);
    mag_max.y = max (mag_max.y, mag.m.y);
    mag_max.z = max (mag_max.z, mag.m.z);

    mag_off.x = mag_min.x + ((mag_max.x - mag_min.x) / 2);
    mag_off.y = mag_min.y + ((mag_max.y - mag_min.y) / 2);
    mag_off.z = mag_min.z + ((mag_max.z - mag_min.z) / 2);
  }
  #else
  mag.m.x = constrain (mag.m.x, mag_min.x, mag_max.x);
  mag.m.y = constrain (mag.m.y, mag_min.y, mag_max.y);
  mag.m.z = constrain (mag.m.z, mag_min.z, mag_max.z);
  
  mgX = (mag.m.x - mag_off.x) / mag_scale;
  mgY = (mag.m.y - mag_off.y) / mag_scale;
  mgZ = (mag.m.z - mag_off.z) / mag_scale;

  az = atan2 (mgZ, -mgY) * rad_to_deg;
  if (az > 0) hdg = az;
  else hdg = 360 + az;
  #endif
}


void Motion_Test () {
  const int16_t NOM_SPEED = 10;
  const int16_t MAX_SPEED = 20;
  const int16_t MIN_SPEED = -20;
  const double Kp = 0.26;
  const double Kd = 0.4;
  double error, speedDifference;
  static uint16_t lastTime;
  static int16_t lastpos;
  int16_t lnpos, L, R;
  uint16_t currentMillis = millis ();
  
  if (currentMillis - lastTime >= 100) {
    // update time limit
    lastTime = currentMillis;
    // Read robot position over the line.
    lnpos = Read_LineSensor ();
    // Clear status variables.
    speedDifference = error = 0.0;
    if (abs(lnpos) <= 60) {
      // Calculate the robot position error
      error = Kd * (lnpos - lastpos);
      // Update the position information
      lastpos = lnpos;
      // Calculate encoder counts.
      speedDifference = Kp * lnpos + error;
      L = NOM_SPEED + speedDifference;
      R = NOM_SPEED - speedDifference;
      L = constrain (L, MIN_SPEED, MAX_SPEED);
      R = constrain (R, MIN_SPEED, MAX_SPEED);
    }
    else if (lnpos > 60) {
      // All sensors on the line
      // Twist clockwise.
      lastpos = 0;
      L = 0;
      R = 15;
    }
    else {
      // Line lost, Twist counter
      // clockwise.
      lastpos = 0;
      L = 15;
      R = 0;
    }
    // Move the robot.
    balanceDrive (L, R);
  }
}

void Test_Track () {
  // define the next timed state.
  static uint8_t Chk;
  // capture state time to compare with
  // processor time in milliseconds.
  static uint16_t lastTime;
  // Drive counts.
  static int16_t Lftspd, Rgtspd;
  // Read the time in milliseconds.
  uint16_t currentMillis = millis ();

  // Turn 180 degrees clockwise.
  if (Chk == 0 && (currentMillis - lastTime >= 2500)) {
    Chk = 1;
    lastTime = currentMillis;
    Lftspd = 18;
    Rgtspd = -18;
  }
  // Drive forward.
  else if (Chk == 1 && (currentMillis - lastTime >= 1000)) {
    Chk = 2;
    lastTime = currentMillis;
    Lftspd = 18;
    Rgtspd = 18;
  }
  // Turn 180 degrees counter clockwise.
  else if (Chk == 2 && (currentMillis - lastTime >= 2500)) {
    Chk = 3;
    lastTime = currentMillis;
    Lftspd = -18;
    Rgtspd = 18;
  }
  // Move forward.
  else if (Chk >= 3 && (currentMillis - lastTime >= 1000)) {
    Chk = 0;
    lastTime = currentMillis;
    Lftspd = 18;
    Rgtspd = 18;
  }
  balanceDrive (Lftspd, Rgtspd);
}

void standUp()
{
  motors.setSpeeds(0, 0);
  ledGreen(true);
  ledRed(true);
  ledYellow(true);
  // Instead of waiting for tune to stop,
  // Wait for release of the button.
  delay (1000);
/*  
  motors.setSpeeds(-MOTOR_SPEED_LIMIT, -MOTOR_SPEED_LIMIT);
  delay(450);
  motors.setSpeeds(250, 250);
  for (uint8_t i = 0; i < 20; i++)
  {
    delay(UPDATE_TIME_MS);
    balanceUpdateSensors();
    if(angle < 60000)
    {
      break;
    }
  }
*/  
  motorSpeed = 0;
  balanceResetEncoders ();
  ledGreen (false);
  ledRed (false);
  ledYellow (false);
}
